@@ -1,7 +1,7 @@
 #!/bin/sh
 #
 # $RoughId: test.sh,v 1.5 2001/07/13 15:38:27 knu Exp $
-# $Id: test.sh,v 1.1.2.1 2001/08/16 07:35:42 knu Exp $
+# $Id: test.sh 11708 2007-02-12 23:01:19Z shyouhei $
 
 RUBY=${RUBY:=ruby}
 MAKE=${MAKE:=make}
@@ -11,17 +11,20 @@ ${RUBY} extconf.rb --with-cflags="${CFLAGS}"
 ${MAKE} clean
 ${MAKE}
 
-mkdir -p lib/digest
-
 for algo in md5 rmd160 sha1 sha2; do
+    args=--with-cflags="${CFLAGS}"
+
+    if [ $WITH_BUNDLED_ENGINES ]; then
+	args="$args --with-bundled-$algo"
+    fi
+
     (cd $algo &&
-	${RUBY} extconf.rb --with-cflags="${CFLAGS}";
+	${RUBY} extconf.rb $args;
 	${MAKE} clean;
 	${MAKE})
     ln -sf ../../$algo/$algo.so lib/digest/
 done
 
-${RUBY} -I. -I./lib test.rb 
+${RUBY} -I. -I./lib ../../test/digest/test_digest.rb
 
 rm lib/digest/*.so
-rmdir lib/digest
