@@ -1,5 +1,5 @@
 /*
- * $Id: ossl_bio.c,v 1.2 2003/09/17 09:05:01 gotoyuzo Exp $
+ * $Id: ossl_bio.c 12496 2007-06-08 15:02:04Z technorama $
  * 'OpenSSL for Ruby' team members
  * Copyright (C) 2003
  * All rights reserved.
@@ -9,6 +9,9 @@
  * (See the file 'LICENCE'.)
  */
 #include "ossl.h"
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 
 BIO *
 ossl_obj2bio(VALUE obj)
@@ -16,16 +19,29 @@ ossl_obj2bio(VALUE obj)
     BIO *bio;
 
     if (TYPE(obj) == T_FILE) {
-	OpenFile *fptr;
+	rb_io_t *fptr;
+	FILE *fp;
+	int fd;
+
 	GetOpenFile(obj, fptr);
 	rb_io_check_readable(fptr);
-	bio = BIO_new_fp(fptr->f, BIO_NOCLOSE);
-    }       
+	if ((fd = dup(FPTR_TO_FD(fptr))) < 0){
+	    rb_sys_fail(0);
+	}
+	if (!(fp = fdopen(fd, "r"))){
+	    close(fd);
+	    rb_sys_fail(0);
+	}
+	if (!(bio = BIO_new_fp(fp, BIO_CLOSE))){
+	    fclose(fp);
+	    ossl_raise(eOSSLError, NULL);
+	}
+    }
     else {
 	StringValue(obj);
-	bio = BIO_new_mem_buf(RSTRING(obj)->ptr, RSTRING(obj)->len);
+	bio = BIO_new_mem_buf(RSTRING_PTR(obj), RSTRING_LEN(obj));
+	if (!bio) ossl_raise(eOSSLError, NULL);
     }
-    if (!bio) ossl_raise(eOSSLError, NULL);
 
     return bio;
 }
