@@ -1,5 +1,5 @@
 /*
- * $Id: ossl_pkey_rsa.c,v 1.5.2.3 2004/12/05 16:43:26 gotoyuzo Exp $
+ * $Id: ossl_pkey_rsa.c 15611 2008-02-26 07:11:23Z technorama $
  * 'OpenSSL for Ruby' project
  * Copyright (C) 2001-2002  Michal Rokos <m.rokos@sh.cvut.cz>
  * All rights reserved.
@@ -20,12 +20,7 @@
 } while (0)
 
 #define RSA_HAS_PRIVATE(rsa) ((rsa)->p && (rsa)->q)
-
-#ifdef OSSL_ENGINE_ENABLED
-#  define RSA_PRIVATE(rsa) (RSA_HAS_PRIVATE(rsa) || (rsa)->engine)
-#else
-#  define RSA_PRIVATE(rsa) RSA_HAS_PRIVATE(rsa)
-#endif
+#define RSA_PRIVATE(obj,rsa) (RSA_HAS_PRIVATE(rsa)||OSSL_PKEY_IS_PRIVATE(obj))
 
 /*
  * Classes
@@ -89,9 +84,19 @@ rsa_generate(int size, int exp)
 	    NULL);
 }
 
+/*
+ *  call-seq:
+ *     RSA.generate(size [, exponent]) -> rsa
+ *
+ *  === Parameters
+ *  * +size+ is an integer representing the desired key size.  Keys smaller than 1024 should be considered insecure.
+ *  * +exponent+ is an odd number normally 3, 17, or 65537.
+ *
+ */
 static VALUE
 ossl_rsa_s_generate(int argc, VALUE *argv, VALUE klass)
 {
+/* why does this method exist?  why can't initialize take an optional exponent? */
     RSA *rsa;
     VALUE size, exp;
     VALUE obj;
@@ -109,6 +114,20 @@ ossl_rsa_s_generate(int argc, VALUE *argv, VALUE klass)
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     RSA.new([size | encoded_key] [, pass]) -> rsa
+ *
+ *  === Parameters
+ *  * +size+ is an integer representing the desired key size.
+ *  * +encoded_key+ is a string containing PEM or DER encoded key.
+ *  * +pass+ is an optional string with the password to decrypt the encoded key.
+ *
+ *  === Examples
+ *  * RSA.new(2048) -> rsa 
+ *  * RSA.new(File.read("rsa.pem")) -> rsa
+ *  * RSA.new(File.read("rsa.pem"), "mypassword") -> rsa
+ */
 static VALUE
 ossl_rsa_initialize(int argc, VALUE *argv, VALUE self)
 {
@@ -162,6 +181,13 @@ ossl_rsa_initialize(int argc, VALUE *argv, VALUE self)
     return self;
 }
 
+/*
+ *  call-seq:
+ *     rsa.public? -> true
+ *
+ *  The return value is always true since every private key is also a public key.
+ *
+ */
 static VALUE
 ossl_rsa_is_public(VALUE self)
 {
@@ -169,22 +195,38 @@ ossl_rsa_is_public(VALUE self)
 
     GetPKeyRSA(self, pkey);
     /*
-     * SURPRISE! :-))
-     * Every key is public at the same time!
+     * This method should check for n and e.  BUG.
      */
     return Qtrue;
 }
 
+/*
+ *  call-seq:
+ *     rsa.private? -> true | false
+ *
+ */
 static VALUE
 ossl_rsa_is_private(VALUE self)
 {
     EVP_PKEY *pkey;
 	
     GetPKeyRSA(self, pkey);
-	
-    return (RSA_PRIVATE(pkey->pkey.rsa)) ? Qtrue : Qfalse;
+    
+    return (RSA_PRIVATE(self, pkey->pkey.rsa)) ? Qtrue : Qfalse;
 }
 
+/*
+ *  call-seq:
+ *     rsa.to_pem([cipher, pass]) -> aString
+ *
+ *  === Parameters
+ *  * +cipher+ is a Cipher object.
+ *  * +pass+ is a string.
+ *
+ *  === Examples
+ *  * rsa.to_pem -> aString
+ *  * rsa.to_pem(cipher, pass) -> aString
+ */
 static VALUE
 ossl_rsa_export(int argc, VALUE *argv, VALUE self)
 {
@@ -224,6 +266,11 @@ ossl_rsa_export(int argc, VALUE *argv, VALUE self)
     return str;
 }
 
+/*
+ *  call-seq:
+ *     rsa.to_der -> aString
+ *
+ */
 static VALUE
 ossl_rsa_to_der(VALUE self)
 {
@@ -241,7 +288,7 @@ ossl_rsa_to_der(VALUE self)
     if((len = i2d_func(pkey->pkey.rsa, NULL)) <= 0)
 	ossl_raise(eRSAError, NULL);
     str = rb_str_new(0, len);
-    p = RSTRING(str)->ptr;
+    p = RSTRING_PTR(str);
     if(i2d_func(pkey->pkey.rsa, &p) < 0)
 	ossl_raise(eRSAError, NULL);
     ossl_str_adjust(str, p);
@@ -251,6 +298,11 @@ ossl_rsa_to_der(VALUE self)
 
 #define ossl_rsa_buf_size(pkey) (RSA_size((pkey)->pkey.rsa)+16)
 
+/*
+ *  call-seq:
+ *     rsa.public_encrypt(string [, padding]) -> aString
+ *
+ */
 static VALUE
 ossl_rsa_public_encrypt(int argc, VALUE *argv, VALUE self)
 {
@@ -263,16 +315,20 @@ ossl_rsa_public_encrypt(int argc, VALUE *argv, VALUE self)
     pad = (argc == 1) ? RSA_PKCS1_PADDING : NUM2INT(padding);
     StringValue(buffer);
     str = rb_str_new(0, ossl_rsa_buf_size(pkey));
-    buf_len = RSA_public_encrypt(RSTRING(buffer)->len, RSTRING(buffer)->ptr,
-				 RSTRING(str)->ptr, pkey->pkey.rsa,
+    buf_len = RSA_public_encrypt(RSTRING_LEN(buffer), RSTRING_PTR(buffer),
+				 RSTRING_PTR(str), pkey->pkey.rsa,
 				 pad);
     if (buf_len < 0) ossl_raise(eRSAError, NULL);
-    RSTRING(str)->len = buf_len;
-    RSTRING(str)->ptr[buf_len] = 0;
+    rb_str_set_len(str, buf_len);
 
     return str;
 }
 
+/*
+ *  call-seq:
+ *     rsa.public_decrypt(string [, padding]) -> aString
+ *
+ */
 static VALUE
 ossl_rsa_public_decrypt(int argc, VALUE *argv, VALUE self)
 {
@@ -285,16 +341,20 @@ ossl_rsa_public_decrypt(int argc, VALUE *argv, VALUE self)
     pad = (argc == 1) ? RSA_PKCS1_PADDING : NUM2INT(padding);
     StringValue(buffer);
     str = rb_str_new(0, ossl_rsa_buf_size(pkey));
-    buf_len = RSA_public_decrypt(RSTRING(buffer)->len, RSTRING(buffer)->ptr,
-				 RSTRING(str)->ptr, pkey->pkey.rsa,
+    buf_len = RSA_public_decrypt(RSTRING_LEN(buffer), RSTRING_PTR(buffer),
+				 RSTRING_PTR(str), pkey->pkey.rsa,
 				 pad);
     if (buf_len < 0) ossl_raise(eRSAError, NULL);
-    RSTRING(str)->len = buf_len;
-    RSTRING(str)->ptr[buf_len] = 0;
+    rb_str_set_len(str, buf_len);
     
     return str;
 }
 
+/*
+ *  call-seq:
+ *     rsa.private_encrypt(string [, padding]) -> aString
+ *
+ */
 static VALUE
 ossl_rsa_private_encrypt(int argc, VALUE *argv, VALUE self)
 {
@@ -303,23 +363,28 @@ ossl_rsa_private_encrypt(int argc, VALUE *argv, VALUE self)
     VALUE str, buffer, padding;
 
     GetPKeyRSA(self, pkey);
-    if (!RSA_PRIVATE(pkey->pkey.rsa)) {
+    if (!RSA_PRIVATE(self, pkey->pkey.rsa)) {
 	ossl_raise(eRSAError, "private key needed.");
     }	
     rb_scan_args(argc, argv, "11", &buffer, &padding);
     pad = (argc == 1) ? RSA_PKCS1_PADDING : NUM2INT(padding);
     StringValue(buffer);
     str = rb_str_new(0, ossl_rsa_buf_size(pkey));
-    buf_len = RSA_private_encrypt(RSTRING(buffer)->len, RSTRING(buffer)->ptr,
-				  RSTRING(str)->ptr, pkey->pkey.rsa,
+    buf_len = RSA_private_encrypt(RSTRING_LEN(buffer), RSTRING_PTR(buffer),
+				  RSTRING_PTR(str), pkey->pkey.rsa,
 				  pad);
     if (buf_len < 0) ossl_raise(eRSAError, NULL);
-    RSTRING(str)->len = buf_len;
-    RSTRING(str)->ptr[buf_len] = 0;
+    rb_str_set_len(str, buf_len);
     
     return str;
 }
 
+
+/*
+ *  call-seq:
+ *     rsa.private_decrypt(string [, padding]) -> aString
+ *
+ */
 static VALUE
 ossl_rsa_private_decrypt(int argc, VALUE *argv, VALUE self)
 {
@@ -328,24 +393,26 @@ ossl_rsa_private_decrypt(int argc, VALUE *argv, VALUE self)
     VALUE str, buffer, padding;
 
     GetPKeyRSA(self, pkey);
-    if (!RSA_PRIVATE(pkey->pkey.rsa)) {
+    if (!RSA_PRIVATE(self, pkey->pkey.rsa)) {
 	ossl_raise(eRSAError, "private key needed.");
     }
     rb_scan_args(argc, argv, "11", &buffer, &padding);
     pad = (argc == 1) ? RSA_PKCS1_PADDING : NUM2INT(padding);
     StringValue(buffer);
     str = rb_str_new(0, ossl_rsa_buf_size(pkey));
-    buf_len = RSA_private_decrypt(RSTRING(buffer)->len, RSTRING(buffer)->ptr,
-				  RSTRING(str)->ptr, pkey->pkey.rsa,
+    buf_len = RSA_private_decrypt(RSTRING_LEN(buffer), RSTRING_PTR(buffer),
+				  RSTRING_PTR(str), pkey->pkey.rsa,
 				  pad);
     if (buf_len < 0) ossl_raise(eRSAError, NULL);
-    RSTRING(str)->len = buf_len;
-    RSTRING(str)->ptr[buf_len] = 0;
+    rb_str_set_len(str, buf_len);
 
     return str;
 }
 
 /*
+ *  call-seq:
+ *     rsa.params -> hash
+ *
  * Stores all parameters of key to the hash
  * INSECURE: PRIVATE INFORMATIONS CAN LEAK OUT!!!
  * Don't use :-)) (I's up to you)
@@ -373,6 +440,9 @@ ossl_rsa_get_params(VALUE self)
 }
 
 /*
+ *  call-seq:
+ *     rsa.to_text -> aString
+ *
  * Prints all parameters of key to buffer
  * INSECURE: PRIVATE INFORMATIONS CAN LEAK OUT!!!
  * Don't use :-)) (It's up to you)
@@ -398,6 +468,9 @@ ossl_rsa_to_text(VALUE self)
 }
 
 /*
+ *  call-seq:
+ *     rsa.public_key -> aRSA
+ *
  * Makes new instance RSA PUBLIC_KEY from PRIVATE_KEY
  */
 static VALUE
@@ -420,7 +493,6 @@ ossl_rsa_to_public_key(VALUE self)
 
 /*
  * TODO: Test me
-extern BN_CTX *ossl_bn_ctx;
 
 static VALUE
 ossl_rsa_blinding_on(VALUE self)
@@ -464,6 +536,11 @@ OSSL_PKEY_BN(rsa, iqmp);
 void
 Init_ossl_rsa()
 {
+#if 0 /* let rdoc know about mOSSL and mPKey */
+    mOSSL = rb_define_module("OpenSSL");
+    mPKey = rb_define_module_under(mOSSL, "PKey");
+#endif
+
     eRSAError = rb_define_class_under(mPKey, "RSAError", ePKeyError);
 
     cRSA = rb_define_class_under(mPKey, "RSA", cPKey);
@@ -508,11 +585,9 @@ Init_ossl_rsa()
 }
 
 #else /* defined NO_RSA */
-#   warning >>> OpenSSL is compiled without RSA support <<<
 void
 Init_ossl_rsa()
 {
-    rb_warning("OpenSSL is compiled without RSA support");
 }
 #endif /* NO_RSA */
 
