@@ -1,3 +1,4 @@
+# -*- coding: euc-jp -*-
 #
 # two image widgets demo (called by 'widget')
 #
@@ -15,8 +16,10 @@ $image1_demo = TkToplevel.new {|w|
   positionWindow(w)
 }
 
+base_frame = TkFrame.new($image1_demo).pack(:fill=>:both, :expand=>true)
+
 # label 
-msg = TkLabel.new($image1_demo) {
+msg = TkLabel.new(base_frame) {
   font $font
   wraplength '4i'
   justify 'left'
@@ -25,9 +28,10 @@ msg = TkLabel.new($image1_demo) {
 msg.pack('side'=>'top')
 
 # frame 
-TkFrame.new($image1_demo) {|frame|
+TkFrame.new(base_frame) {|frame|
   TkButton.new(frame) {
-    text 'λ'
+    #text 'λ'
+    text 'Ĥ'
     command proc{
       tmppath = $image1_demo
       $image1_demo = nil
@@ -44,14 +48,17 @@ TkFrame.new($image1_demo) {|frame|
 
 # image 
 image1a = \
-TkPhotoImage.new('file'=>[$demo_dir,
+TkPhotoImage.new('file'=>[$demo_dir,'..',
                           'images','earth.gif'].join(File::Separator))
 image1b = \
-TkPhotoImage.new('file'=>[$demo_dir,
+TkPhotoImage.new('file'=>[$demo_dir,'..',
                           'images','earthris.gif'].join(File::Separator))
 
 # label 
-[ TkLabel.new($image1_demo, 'image'=>image1a, 'bd'=>1, 'relief'=>'sunken'),
-  TkLabel.new($image1_demo, 'image'=>image1b, 'bd'=>1, 'relief'=>'sunken')
+#[ TkLabel.new(base_frame, 'image'=>image1a, 'bd'=>1, 'relief'=>'sunken'),
+#  TkLabel.new(base_frame, 'image'=>image1b, 'bd'=>1, 'relief'=>'sunken')
+#].each{|w| w.pack('side'=>'top', 'padx'=>'.5m', 'pady'=>'.5m')}
+[ Tk::Label.new(base_frame, 'image'=>image1a, 'bd'=>1, 'relief'=>'sunken'),
+  Tk::Label.new(base_frame, 'image'=>image1b, 'bd'=>1, 'relief'=>'sunken')
 ].each{|w| w.pack('side'=>'top', 'padx'=>'.5m', 'pady'=>'.5m')}
 
