@@ -2,8 +2,8 @@
 
   marshal.c -
 
-  $Author: akr $
-  $Date: 2005/12/14 03:04:14 $
+  $Author: shyouhei $
+  $Date: 2008-07-03 20:14:50 +0900 (Thu, 03 Jul 2008) $
   created at: Thu Apr 27 16:30:01 JST 1995
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -81,9 +81,19 @@ shortlen(len, ds)
 #define TYPE_LINK	'@'
 
 static ID s_dump, s_load, s_mdump, s_mload;
-static ID s_dump_data, s_load_data, s_alloc;
+static ID s_dump_data, s_load_data, s_alloc, s_call;
 static ID s_getc, s_read, s_write, s_binmode;
 
+static void
+reentrant_check(obj, sym)
+    VALUE obj;
+    ID sym;
+{
+    if (obj && RBASIC(obj)->klass) {
+        rb_raise(rb_eRuntimeError, "%s reentered", rb_id2name(sym));
+    }
+}
+
 struct dump_arg {
     VALUE obj;
     VALUE str, dest;
@@ -120,7 +130,7 @@ static void w_long _((long, struct dump_arg*));
 
 static void
 w_nbyte(s, n, arg)
-    char *s;
+    const char *s;
     int n;
     struct dump_arg *arg;
 {
@@ -143,7 +153,7 @@ w_byte(c, arg)
 
 static void
 w_bytes(s, n, arg)
-    char *s;
+    const char *s;
     int n;
     struct dump_arg *arg;
 {
@@ -328,7 +338,7 @@ w_symbol(id, arg)
     ID id;
     struct dump_arg *arg;
 {
-    char *sym = rb_id2name(id);
+    const char *sym = rb_id2name(id);
     st_data_t num;
 
     if (st_lookup(arg->symbols, id, &num)) {
@@ -344,7 +354,7 @@ w_symbol(id, arg)
 
 static void
 w_unique(s, arg)
-    char *s;
+    const char *s;
     struct dump_arg *arg;
 {
     if (s[0] == '#') {
@@ -371,10 +381,10 @@ w_extended(klass, arg, check)
     struct dump_arg *arg;
     int check;
 {
-    char *path;
+    const char *path;
 
-    if (FL_TEST(klass, FL_SINGLETON)) {
-	if (check && RCLASS(klass)->m_tbl->num_entries ||
+    if (check && FL_TEST(klass, FL_SINGLETON)) {
+	if (RCLASS(klass)->m_tbl->num_entries ||
 	    (RCLASS(klass)->iv_tbl && RCLASS(klass)->iv_tbl->num_entries > 1)) {
 	    rb_raise(rb_eTypeError, "singleton can't be dumped");
 	}
@@ -468,7 +478,7 @@ w_object(obj, arg, limit)
 	return;
     }
 
-    if (ivtbl = rb_generic_ivar_table(obj)) {
+    if ((ivtbl = rb_generic_ivar_table(obj)) != 0) {
 	w_byte(TYPE_IVAR, arg);
     }
     if (obj == Qnil) {
@@ -502,9 +512,10 @@ w_object(obj, arg, limit)
 
 	st_add_direct(arg->data, obj, arg->data->num_entries);
 	if (rb_respond_to(obj, s_mdump)) {
-	    VALUE v;
+	    volatile VALUE v;
 
 	    v = rb_funcall(obj, s_mdump, 0, 0);
+	    reentrant_check(arg->str, s_mdump);
 	    w_class(TYPE_USRMARSHAL, obj, arg, Qfalse);
 	    w_object(v, arg, limit);
 	    if (ivtbl) w_ivar(0, &c_arg);
@@ -514,6 +525,7 @@ w_object(obj, arg, limit)
 	    VALUE v;
 
 	    v = rb_funcall(obj, s_dump, 1, INT2NUM(limit));
+	    reentrant_check(arg->str, s_dump);
 	    if (TYPE(v) != T_STRING) {
 		rb_raise(rb_eTypeError, "_dump() must return string");
 	    }
@@ -652,13 +664,14 @@ w_object(obj, arg, limit)
 	    {
 		VALUE v;
 
-		w_class(TYPE_DATA, obj, arg, Qtrue);
 		if (!rb_respond_to(obj, s_dump_data)) {
 		    rb_raise(rb_eTypeError,
 			     "no marshal_dump is defined for class %s",
 			     rb_obj_classname(obj));
 		}
 		v = rb_funcall(obj, s_dump_data, 0);
+		reentrant_check(arg->str, s_dump_data);
+		w_class(TYPE_DATA, obj, arg, Qtrue);
 		w_object(v, arg, limit);
 	    }
 	    break;
@@ -690,11 +703,13 @@ static VALUE
 dump_ensure(arg)
     struct dump_arg *arg;
 {
+    if (RBASIC(arg->str)->klass) return 0; /* ignore reentrant */
     st_free_table(arg->symbols);
     st_free_table(arg->data);
     if (arg->taint) {
 	OBJ_TAINT(arg->str);
     }
+
     return 0;
 }
 
@@ -747,20 +762,21 @@ marshal_dump(argc, argv)
 	else port = a1;
     }
     arg.dest = 0;
+    arg.str = rb_str_buf_new(0);
+    RBASIC(arg.str)->klass = 0;
     if (!NIL_P(port)) {
 	if (!rb_respond_to(port, s_write)) {
 	  type_error:
 	    rb_raise(rb_eTypeError, "instance of IO needed");
 	}
-	arg.str = rb_str_buf_new(0);
 	arg.dest = port;
 	if (rb_respond_to(port, s_binmode)) {
 	    rb_funcall2(port, s_binmode, 0, 0);
+	    reentrant_check(arg.str, s_dump_data);
 	}
     }
     else {
-	port = rb_str_buf_new(0);
-	arg.str = port;
+	port = arg.str;
     }
 
     arg.symbols = st_init_numtable();
@@ -774,6 +790,7 @@ marshal_dump(argc, argv)
     w_byte(MARSHAL_MINOR, &arg);
 
     rb_ensure(dump, (VALUE)&c_arg, dump_ensure, (VALUE)&arg);
+    RBASIC(arg.str)->klass = rb_cString;
 
     return port;
 }
@@ -806,6 +823,7 @@ r_byte(arg)
     else {
 	VALUE src = arg->src;
 	VALUE v = rb_funcall2(src, s_getc, 0, 0);
+	reentrant_check(arg->data, s_getc);
 	if (NIL_P(v)) rb_eof_error();
 	c = (unsigned char)FIX2INT(v);
     }
@@ -873,7 +891,7 @@ r_bytes0(len, arg)
 
     if (len == 0) return rb_str_new(0, 0);
     if (TYPE(arg->src) == T_STRING) {
-	if (RSTRING(arg->src)->len > arg->offset) {
+	if (RSTRING(arg->src)->len - arg->offset >= len) {
 	    str = rb_str_new(RSTRING(arg->src)->ptr+arg->offset, len);
 	    arg->offset += len;
 	}
@@ -886,6 +904,7 @@ r_bytes0(len, arg)
 	VALUE src = arg->src;
 	VALUE n = LONG2NUM(len);
 	str = rb_funcall2(src, s_read, 1, &n);
+	reentrant_check(arg->data, s_read);
 	if (NIL_P(str)) goto too_short;
 	StringValue(str);
 	if (RSTRING(str)->len != len) goto too_short;
@@ -929,7 +948,7 @@ r_symbol(arg)
     return r_symreal(arg);
 }
 
-static char*
+static const char*
 r_unique(arg)
     struct load_arg *arg;
 {
@@ -972,7 +991,7 @@ r_ivar(obj, arg)
 
 static VALUE
 path2class(path)
-    char *path;
+    const char *path;
 {
     VALUE v = rb_path2class(path);
 
@@ -984,7 +1003,7 @@ path2class(path)
 
 static VALUE
 path2module(path)
-    char *path;
+    const char *path;
 {
     VALUE v = rb_path2class(path);
 
@@ -1199,9 +1218,6 @@ r_object0(arg, proc, ivp, extmod)
 
 	    klass = path2class(r_unique(arg));
 	    mem = rb_struct_s_members(klass);
-	    if (mem == Qnil) {
-		rb_raise(rb_eTypeError, "uninitialized struct");
-	    }
 	    len = r_long(arg);
 
 	    values = rb_ary_new2(len);
@@ -1239,6 +1255,7 @@ r_object0(arg, proc, ivp, extmod)
 		*ivp = Qfalse;
 	    }
 	    v = rb_funcall(klass, s_load, 1, data);
+	    reentrant_check(arg->data, s_load);
 	    r_entry(v, arg);
 	}
         break;
@@ -1262,6 +1279,7 @@ r_object0(arg, proc, ivp, extmod)
 	    r_entry(v, arg);
 	    data = r_object(arg);
 	    rb_funcall(v, s_mload, 1, data);
+	    reentrant_check(arg->data, s_mload);
 	}
         break;
 
@@ -1288,6 +1306,7 @@ r_object0(arg, proc, ivp, extmod)
 		   warn = Qfalse;
 	       }
 	       v = rb_funcall(klass, s_alloc, 0);
+	       reentrant_check(arg->data, s_alloc);
            }
 	   else {
 	       v = rb_obj_alloc(klass);
@@ -1302,6 +1321,7 @@ r_object0(arg, proc, ivp, extmod)
                         rb_class2name(klass));
            }
            rb_funcall(v, s_load_data, 1, r_object0(arg, 0, 0, extmod));
+	   reentrant_check(arg->data, s_load_data);
        }
        break;
 
@@ -1344,7 +1364,8 @@ r_object0(arg, proc, ivp, extmod)
 	break;
     }
     if (proc) {
-	rb_funcall(proc, rb_intern("call"), 1, v);
+	rb_funcall(proc, s_call, 1, v);
+	reentrant_check(arg->data, s_call);
     }
     return v;
 }
@@ -1367,6 +1388,7 @@ static VALUE
 load_ensure(arg)
     struct load_arg *arg;
 {
+    if (RBASIC(arg->data)->klass) return 0; /* ignore reentrant */
     st_free_table(arg->symbols);
     return 0;
 }
@@ -1393,9 +1415,10 @@ marshal_load(argc, argv)
     struct load_arg arg;
 
     rb_scan_args(argc, argv, "11", &port, &proc);
-    if (rb_respond_to(port, rb_intern("to_str"))) {
+    v = rb_check_string_type(port);
+    if (!NIL_P(v)) {
 	arg.taint = OBJ_TAINTED(port); /* original taintedness */
-	StringValue(port);	       /* possible conversion */
+	port = v;
     }
     else if (rb_respond_to(port, s_getc) && rb_respond_to(port, s_read)) {
 	if (rb_respond_to(port, s_binmode)) {
@@ -1408,6 +1431,7 @@ marshal_load(argc, argv)
     }
     arg.src = port;
     arg.offset = 0;
+    arg.data = 0;
 
     major = r_byte(&arg);
     minor = r_byte(&arg);
@@ -1424,9 +1448,11 @@ marshal_load(argc, argv)
 
     arg.symbols = st_init_numtable();
     arg.data   = rb_hash_new();
+    RBASIC(arg.data)->klass = 0;
     if (NIL_P(proc)) arg.proc = 0;
     else             arg.proc = proc;
     v = rb_ensure(load, (VALUE)&arg, load_ensure, (VALUE)&arg);
+    RBASIC(arg.data)->klass = rb_cHash;
 
     return v;
 }
@@ -1476,6 +1502,7 @@ Init_marshal()
     s_dump_data = rb_intern("_dump_data");
     s_load_data = rb_intern("_load_data");
     s_alloc = rb_intern("_alloc");
+    s_call = rb_intern("call");
     s_getc = rb_intern("getc");
     s_read = rb_intern("read");
     s_write = rb_intern("write");
