@@ -2,8 +2,8 @@
 
   signal.c -
 
-  $Author: nagai $
-  $Date: 2006/07/10 05:01:57 $
+  $Author: knu $
+  $Date: 2008-06-06 19:39:57 +0900 (Fri, 06 Jun 2008) $
   created at: Tue Dec 20 10:13:44 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -21,6 +21,12 @@
 #undef SIGBUS
 #endif
 
+#if defined HAVE_SIGPROCMASK || defined HAVE_SIGSETMASK
+#define USE_TRAP_MASK 1
+#else
+#define USE_TRAP_MASK 0
+#endif
+
 #ifndef NSIG
 # ifdef DJGPP
 #  define NSIG SIGMAX
@@ -30,7 +36,7 @@
 #endif
 
 static struct signals {
-    char *signm;
+    const char *signm;
     int  signo;
 } siglist [] = {
     {"EXIT", 0},
@@ -170,7 +176,7 @@ static struct signals {
 
 static int
 signm2signo(nm)
-    char *nm;
+    const char *nm;
 {
     struct signals *sigs;
 
@@ -180,7 +186,7 @@ signm2signo(nm)
     return 0;
 }
 
-static char*
+static const char*
 signo2signm(no)
     int no;
 {
@@ -199,6 +205,96 @@ ruby_signal_name(no)
     return signo2signm(no);
 }
 
+/*
+ * call-seq:
+ *    SignalException.new(sig)   =>  signal_exception
+ *
+ *  Construct a new SignalException object.  +sig+ should be a known
+ *  signal name, or a signal number.
+ */
+
+static VALUE
+esignal_init(argc, argv, self)
+    int argc;
+    VALUE *argv;
+    VALUE self;
+{
+    int argnum = 1;
+    VALUE sig = Qnil;
+    int signo;
+    const char *signm;
+    char tmpnm[(sizeof(int)*CHAR_BIT)/3+4];
+
+    if (argc > 0) {
+	sig = argv[0];
+	if (FIXNUM_P(sig)) argnum = 2;
+    }
+    if (argc < 1 || argnum < argc) {
+	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
+		 argc, argnum);
+    }
+    if (argnum == 2) {
+	signo = FIX2INT(sig);
+	if (signo < 0 || signo > NSIG) {
+	    rb_raise(rb_eArgError, "invalid signal number (%d)", signo);
+	}
+	if (argc > 1) {
+	    sig = argv[1];
+	}
+	else {
+	    signm = signo2signm(signo);
+	    if (signm) {
+		snprintf(tmpnm, sizeof(tmpnm), "SIG%s", signm);
+	    }
+	    else {
+		snprintf(tmpnm, sizeof(tmpnm), "SIG%u", signo);
+	    }
+	    sig = rb_str_new2(signm = tmpnm);
+	}
+    }
+    else {
+	signm = SYMBOL_P(sig) ? rb_id2name(SYM2ID(sig)) : StringValuePtr(sig);
+	if (strncmp(signm, "SIG", 3) == 0) signm += 3;
+	signo = signm2signo(signm);
+	if (!signo) {
+	    rb_raise(rb_eArgError, "unsupported name `SIG%s'", signm);
+	}
+	if (SYMBOL_P(sig)) {
+	    sig = rb_str_new2(signm);
+	}
+    }
+    rb_call_super(1, &sig);
+    rb_iv_set(self, "signo", INT2NUM(signo));
+
+    return self;
+}
+
+static VALUE
+interrupt_init(argc, argv, self)
+    int argc;
+    VALUE *argv;
+    VALUE self;
+{
+    VALUE args[2];
+
+    args[0] = INT2FIX(SIGINT);
+    rb_scan_args(argc, argv, "01", &args[1]);
+
+    return rb_call_super(2, args);
+}
+
+void
+ruby_default_signal(sig)
+    int sig;
+{
+#ifndef MACOS_UNUSE_SIGNAL
+    extern rb_pid_t getpid _((void));
+
+    signal(sig, SIG_DFL);
+    kill(getpid(), sig);
+#endif
+}
+
 /*
  *  call-seq:
  *     Process.kill(signal, pid, ...)    => fixnum
@@ -231,7 +327,7 @@ rb_f_kill(argc, argv)
     int negative = 0;
     int sig;
     int i;
-    char *s;
+    const char *s;
 
     rb_secure(2);
     if (argc < 2)
@@ -341,14 +437,6 @@ ruby_signal(signum, handler)
     sigact.sa_handler = handler;
     sigemptyset(&sigact.sa_mask);
     sigact.sa_flags = 0;
-# if defined(SA_RESTART)
-    /* All other signals but VTALRM shall restart restartable syscall
-       VTALRM will cause EINTR to syscall if interrupted.
-    */
-    if (signum != SIGVTALRM) {
-        sigact.sa_flags |= SA_RESTART; /* SVR4, 4.3+BSD */
-    }
-# endif
 # ifdef SA_NOCLDWAIT
     if (signum == SIGCHLD && handler == SIG_IGN)
 	sigact.sa_flags |= SA_NOCLDWAIT;
@@ -421,6 +509,9 @@ signal_exec(sig)
 #ifdef SIGQUIT
 	  case SIGQUIT:
 #endif
+#ifdef SIGTERM
+	  case SIGTERM:
+#endif
 #ifdef SIGALRM
 	  case SIGALRM:
 #endif
@@ -430,7 +521,7 @@ signal_exec(sig)
 #ifdef SIGUSR2
 	  case SIGUSR2:
 #endif
-	    rb_thread_signal_raise(signo2signm(sig));
+	    rb_thread_signal_raise(sig);
 	    break;
 	}
     }
@@ -481,16 +572,17 @@ sighandler(sig)
 
 #if defined(HAVE_NATIVETHREAD) && defined(HAVE_NATIVETHREAD_KILL)
     if (!is_ruby_native_thread() && !rb_trap_accept_nativethreads[sig]) {
-        sigsend_to_ruby_thread(sig);
-        return;
+	sigsend_to_ruby_thread(sig);
+	return;
     }
 #endif
 
 #if !defined(BSD_SIGNAL) && !defined(POSIX_SIGNAL)
     if (rb_trap_accept_nativethreads[sig]) {
-        ruby_nativethread_signal(sig, sighandler);
-    } else {
-        ruby_signal(sig, sighandler);
+	ruby_nativethread_signal(sig, sighandler);
+    }
+    else {
+	ruby_signal(sig, sighandler);
     }
 #endif
 
@@ -501,6 +593,9 @@ sighandler(sig)
     else {
 	ATOMIC_INC(rb_trap_pending);
 	ATOMIC_INC(trap_pending_list[sig]);
+#ifdef _WIN32
+	rb_w32_interrupted();
+#endif
     }
 }
 
@@ -534,6 +629,8 @@ sigsegv(sig)
     }
 #endif
 
+    extern int ruby_gc_stress;
+    ruby_gc_stress = 0;
     rb_bug("Segmentation fault");
 }
 #endif
@@ -578,7 +675,7 @@ rb_trap_exec()
 }
 
 struct trap_arg {
-#ifndef _WIN32
+#if USE_TRAP_MASK
 # ifdef HAVE_SIGPROCMASK
     sigset_t mask;
 # else
@@ -588,11 +685,13 @@ struct trap_arg {
     VALUE sig, cmd;
 };
 
+#if USE_TRAP_MASK
 # ifdef HAVE_SIGPROCMASK
 static sigset_t trap_last_mask;
 # else
 static int trap_last_mask;
 # endif
+#endif
 
 static RETSIGTYPE sigexit _((int));
 static RETSIGTYPE
@@ -609,7 +708,7 @@ trap(arg)
     sighandler_t func, oldfunc;
     VALUE command, oldcmd;
     int sig = -1;
-    char *s;
+    const char *s;
 
     func = sighandler;
     command = arg->cmd;
@@ -685,6 +784,9 @@ trap(arg)
 #ifdef SIGQUIT
 	  case SIGQUIT:
 #endif
+#ifdef SIGTERM
+	  case SIGTERM:
+#endif
 #ifdef SIGALRM
 	  case SIGALRM:
 #endif
@@ -724,7 +826,7 @@ trap(arg)
     trap_list[sig].cmd = command;
     trap_list[sig].safe = ruby_safe_level;
     /* enable at least specified signal. */
-#ifndef _WIN32
+#if USE_TRAP_MASK
 #ifdef HAVE_SIGPROCMASK
     sigdelset(&arg->mask, sig);
 #else
@@ -734,7 +836,7 @@ trap(arg)
     return oldcmd;
 }
 
-#ifndef _WIN32
+#if USE_TRAP_MASK
 static VALUE
 trap_ensure(arg)
     struct trap_arg *arg;
@@ -753,7 +855,7 @@ trap_ensure(arg)
 void
 rb_trap_restore_mask()
 {
-#ifndef _WIN32
+#if USE_TRAP_MASK
 # ifdef HAVE_SIGPROCMASK
     sigprocmask(SIG_SETMASK, &trap_last_mask, NULL);
 # else
@@ -813,7 +915,7 @@ sig_trap(argc, argv)
     if (OBJ_TAINTED(arg.cmd)) {
 	rb_raise(rb_eSecurityError, "Insecure: tainted signal trap");
     }
-#ifndef _WIN32
+#if USE_TRAP_MASK
     /* disable interrupt */
 # ifdef HAVE_SIGPROCMASK
     sigfillset(&arg.mask);
@@ -891,12 +993,13 @@ install_nativethread_sighandler(signum, handler)
 #endif
 #endif
 
+#if defined(SIGCLD) || defined(SIGCHLD)
 static void
 init_sigchld(sig)
     int sig;
 {
     sighandler_t oldfunc;
-#ifndef _WIN32
+#if USE_TRAP_MASK
 # ifdef HAVE_SIGPROCMASK
     sigset_t mask;
 # else
@@ -904,7 +1007,7 @@ init_sigchld(sig)
 # endif
 #endif
 
-#ifndef _WIN32
+#if USE_TRAP_MASK
     /* disable interrupt */
 # ifdef HAVE_SIGPROCMASK
     sigfillset(&mask);
@@ -921,7 +1024,7 @@ init_sigchld(sig)
 	trap_list[sig].cmd = 0;
     }
 
-#ifndef _WIN32
+#if USE_TRAP_MASK
 #ifdef HAVE_SIGPROCMASK
     sigdelset(&mask, sig);
     sigprocmask(SIG_SETMASK, &mask, NULL);
@@ -932,6 +1035,7 @@ init_sigchld(sig)
     trap_last_mask = mask;
 #endif
 }
+#endif
 
 /*
  * Many operating systems allow signals to be sent to running
@@ -980,6 +1084,11 @@ Init_signal()
     rb_define_module_function(mSignal, "trap", sig_trap, -1);
     rb_define_module_function(mSignal, "list", sig_list, 0);
 
+    rb_define_method(rb_eSignal, "initialize", esignal_init, -1);
+    rb_attr(rb_eSignal, rb_intern("signo"), 1, 0, 0);
+    rb_alias(rb_eSignal, rb_intern("signm"), rb_intern("message"));
+    rb_define_method(rb_eInterrupt, "initialize", interrupt_init, -1);
+
     install_sighandler(SIGINT, sighandler);
 #ifdef SIGHUP
     install_sighandler(SIGHUP, sighandler);
@@ -987,6 +1096,9 @@ Init_signal()
 #ifdef SIGQUIT
     install_sighandler(SIGQUIT, sighandler);
 #endif
+#ifdef SIGTERM
+    install_sighandler(SIGTERM, sighandler);
+#endif
 #ifdef SIGALRM
     install_sighandler(SIGALRM, sighandler);
 #endif
@@ -1007,10 +1119,9 @@ Init_signal()
     install_sighandler(SIGPIPE, sigpipe);
 #endif
 
-#ifdef SIGCLD
+#if defined(SIGCLD)
     init_sigchld(SIGCLD);
-#endif
-#ifdef SIGCHLD
+#elif defined(SIGCHLD)
     init_sigchld(SIGCHLD);
 #endif
 
