@@ -2,8 +2,8 @@
 
   variable.c -
 
-  $Author: matz $
-  $Date: 2006/07/20 07:04:17 $
+  $Author: shyouhei $
+  $Date: 2008-06-29 18:28:00 +0900 (Sun, 29 Jun 2008) $
   created at: Tue Apr 19 23:55:15 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -195,7 +195,7 @@ rb_class_path(klass)
 	return path;
     }
     else {
-	char *s = "Class";
+	const char *s = "Class";
 	size_t len;
 
 	if (TYPE(klass) == T_MODULE) {
@@ -241,16 +241,21 @@ rb_path2class(path)
     const char *pbeg, *p;
     ID id;
     VALUE c = rb_cObject;
+    VALUE str = 0;
 
     if (path[0] == '#') {
 	rb_raise(rb_eArgError, "can't retrieve anonymous class %s", path);
     }
     pbeg = p = path;
     while (*p) {
-	VALUE str;
-
 	while (*p && *p != ':') p++;
-	str = rb_str_new(pbeg, p-pbeg);
+	if (str) {
+	    RSTRING(str)->len = 0;
+	    rb_str_cat(str, pbeg, p-pbeg);
+	}
+	else {
+	    str = rb_str_new(pbeg, p-pbeg);
+	}
 	id = rb_intern(RSTRING(str)->ptr);
 	if (p[0] == ':') {
 	    if (p[1] != ':') goto undefined_class;
@@ -289,14 +294,14 @@ rb_class_name(klass)
     return rb_class_path(rb_class_real(klass));
 }
 
-char *
+const char *
 rb_class2name(klass)
     VALUE klass;
 {
     return RSTRING(rb_class_name(klass))->ptr;
 }
 
-char *
+const char *
 rb_obj_classname(obj)
     VALUE obj;
 {
@@ -342,8 +347,9 @@ rb_global_entry(id)
     ID id;
 {
     struct global_entry *entry;
+    st_data_t data;
 
-    if (!st_lookup(rb_global_tbl, id, (st_data_t *)&entry)) {
+    if (!st_lookup(rb_global_tbl, id, &data)) {
 	struct global_variable *var;
 	entry = ALLOC(struct global_entry);
 	var = ALLOC(struct global_variable);
@@ -359,6 +365,9 @@ rb_global_entry(id)
 	var->trace = 0;
 	st_add_direct(rb_global_tbl, id, (st_data_t)entry);
     }
+    else {
+	entry = (struct global_entry *)data;
+    }
     return entry;
 }
 
@@ -636,14 +645,16 @@ rb_f_untrace_var(argc, argv)
     ID id;
     struct global_entry *entry;
     struct trace_var *trace;
+    st_data_t data;
 
+    rb_secure(4);
     rb_scan_args(argc, argv, "11", &var, &cmd);
     id = rb_to_id(var);
-    if (!st_lookup(rb_global_tbl, id, (st_data_t *)&entry)) {
+    if (!st_lookup(rb_global_tbl, id, &data)) {
 	rb_name_error(id, "undefined global variable %s", rb_id2name(id));
     }
 
-    trace = entry->var->trace;
+    trace = (entry = (struct global_entry *)data)->var->trace;
     if (NIL_P(cmd)) {
 	VALUE ary = rb_ary_new();
 
@@ -779,7 +790,7 @@ rb_f_global_variables()
 {
     VALUE ary = rb_ary_new();
     char buf[4];
-    char *s = "&`'+123456789";
+    const char *s = "&`'+123456789";
 
     st_foreach(rb_global_tbl, gvar_i, ary);
     if (!NIL_P(rb_backref_get())) {
@@ -797,17 +808,18 @@ rb_alias_variable(name1, name2)
     ID name2;
 {
     struct global_entry *entry1, *entry2;
+    st_data_t data1;
 
     if (rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't alias global variable");
 
     entry2 = rb_global_entry(name2);
-    if (!st_lookup(rb_global_tbl, name1, (st_data_t *)&entry1)) {
+    if (!st_lookup(rb_global_tbl, name1, &data1)) {
 	entry1 = ALLOC(struct global_entry);
 	entry1->id = name1;
 	st_add_direct(rb_global_tbl, name1, (st_data_t)entry1);
     }
-    else if (entry1->var != entry2->var) {
+    else if ((entry1 = (struct global_entry *)data1)->var != entry2->var) {
 	struct global_variable *var = entry1->var;
 	if (var->block_trace) {
 	    rb_raise(rb_eRuntimeError, "can't alias in tracer");
@@ -837,12 +849,12 @@ st_table*
 rb_generic_ivar_table(obj)
     VALUE obj;
 {
-    st_table *tbl;
+    st_data_t tbl;
 
     if (!FL_TEST(obj, FL_EXIVAR)) return 0;
     if (!generic_iv_tbl) return 0;
-    if (!st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) return 0;
-    return tbl;
+    if (!st_lookup(generic_iv_tbl, obj, &tbl)) return 0;
+    return (st_table *)tbl;
 }
 
 static VALUE
@@ -851,15 +863,15 @@ generic_ivar_get(obj, id, warn)
     ID id;
     int warn;
 {
-    st_table *tbl;
+    st_data_t tbl;
     VALUE val;
 
     if (generic_iv_tbl) {
-      if (st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) {
-	if (st_lookup(tbl, id, &val)) {
-	  return val;
+	if (st_lookup(generic_iv_tbl, obj, &tbl)) {
+	    if (st_lookup((st_table *)tbl, id, &val)) {
+		return val;
+	    }
 	}
-      }
     }
     if (warn) {
 	rb_warning("instance variable %s not initialized", rb_id2name(id));
@@ -874,6 +886,7 @@ generic_ivar_set(obj, id, val)
     VALUE val;
 {
     st_table *tbl;
+    st_data_t data;
 
     if (rb_special_const_p(obj)) {
 	special_generic_ivar = 1;
@@ -882,14 +895,14 @@ generic_ivar_set(obj, id, val)
 	generic_iv_tbl = st_init_numtable();
     }
 
-    if (!st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) {
+    if (!st_lookup(generic_iv_tbl, obj, &data)) {
 	FL_SET(obj, FL_EXIVAR);
 	tbl = st_init_numtable();
 	st_add_direct(generic_iv_tbl, obj, (st_data_t)tbl);
 	st_add_direct(tbl, id, val);
 	return;
     }
-    st_insert(tbl, id, val);
+    st_insert((st_table *)data, id, val);
 }
 
 static VALUE
@@ -898,10 +911,12 @@ generic_ivar_defined(obj, id)
     ID id;
 {
     st_table *tbl;
+    st_data_t data;
     VALUE val;
 
     if (!generic_iv_tbl) return Qfalse;
-    if (!st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) return Qfalse;
+    if (!st_lookup(generic_iv_tbl, obj, &data)) return Qfalse;
+    tbl = (st_table *)data;
     if (st_lookup(tbl, id, &val)) {
 	return Qtrue;
     }
@@ -915,14 +930,16 @@ generic_ivar_remove(obj, id, valp)
     VALUE *valp;
 {
     st_table *tbl;
+    st_data_t data;
     int status;
 
     if (!generic_iv_tbl) return 0;
-    if (!st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) return 0;
+    if (!st_lookup(generic_iv_tbl, obj, &data)) return 0;
+    tbl = (st_table *)data;
     status = st_delete(tbl, &id, valp);
     if (tbl->num_entries == 0) {
-	st_delete(generic_iv_tbl, &obj, (st_data_t *)&tbl);
-	st_free_table(tbl);
+	st_delete(generic_iv_tbl, &obj, &data);
+	st_free_table((st_table *)data);
     }
     return status;
 }
@@ -931,11 +948,11 @@ void
 rb_mark_generic_ivar(obj)
     VALUE obj;
 {
-    st_table *tbl;
+    st_data_t tbl;
 
     if (!generic_iv_tbl) return;
-    if (st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) {
-	rb_mark_tbl(tbl);
+    if (st_lookup(generic_iv_tbl, obj, &tbl)) {
+	rb_mark_tbl((st_table *)tbl);
     }
 }
 
@@ -971,24 +988,26 @@ void
 rb_free_generic_ivar(obj)
     VALUE obj;
 {
-    st_table *tbl;
+    st_data_t tbl;
 
-    if (st_delete(generic_iv_tbl, &obj, (st_data_t *)&tbl))
-	st_free_table(tbl);
+    if (!generic_iv_tbl) return;
+    if (st_delete(generic_iv_tbl, &obj, &tbl))
+	st_free_table((st_table *)tbl);
 }
 
 void
 rb_copy_generic_ivar(clone, obj)
     VALUE clone, obj;
 {
-    st_table *tbl;
+    st_data_t data;
 
     if (!generic_iv_tbl) return;
-    if (st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) {
-	st_table *old;
+    if (!FL_TEST(obj, FL_EXIVAR)) return;
+    if (st_lookup(generic_iv_tbl, obj, &data)) {
+	st_table *tbl = (st_table *)data;
 
-	if (st_lookup(generic_iv_tbl, clone, (st_data_t *)&old)) {
-	    st_free_table(old);
+	if (st_lookup(generic_iv_tbl, clone, &data)) {
+	    st_free_table((st_table *)data);
 	    st_insert(generic_iv_tbl, clone, (st_data_t)st_copy(tbl));
 	}
 	else {
@@ -1129,10 +1148,10 @@ rb_obj_instance_variables(obj)
       default:
 	if (!generic_iv_tbl) break;
 	if (FL_TEST(obj, FL_EXIVAR) || rb_special_const_p(obj)) {
-	    st_table *tbl;
+	    st_data_t tbl;
 
-	    if (st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) {
-		st_foreach_safe(tbl, ivar_i, ary);
+	    if (st_lookup(generic_iv_tbl, obj, &tbl)) {
+		st_foreach_safe((st_table *)tbl, ivar_i, ary);
 	    }
 	}
 	break;
@@ -1635,7 +1654,7 @@ mod_av_set(klass, id, val, isconst)
     VALUE val;
     int isconst;
 {
-    char *dest = isconst ? "constant" : "class variable";
+    const char *dest = isconst ? "constant" : "class variable";
 
     if (!OBJ_TAINTED(klass) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't set %s", dest);
@@ -1686,7 +1705,7 @@ rb_define_const(klass, name, val)
     ID id = rb_intern(name);
 
     if (!rb_is_const_id(id)) {
-	rb_warn("rb_define_const: invalide name `%s' for constant", name);
+	rb_warn("rb_define_const: invalid name `%s' for constant", name);
     }
     if (klass == rb_cObject) {
 	rb_secure(4);
