@@ -4,7 +4,9 @@ require 'rbconfig'
 
 CONFIG = Config::MAKEFILE_CONFIG
 
-fversion = "#{CONFIG['MAJOR']},#{CONFIG['MINOR']},#{CONFIG['TEENY']},0"
+version = %w'MAJOR MINOR TEENY PATCHLEVEL'.map {|v| CONFIG[v] || '0'}
+fversion = version.join(',')
+rversion = version.join('.')
 
 $ruby_name ||= CONFIG["RUBY_INSTALL_NAME"]
 $rubyw_name ||= CONFIG["RUBYW_INSTALL_NAME"] || $ruby_name.sub(/ruby/, '\&w')
@@ -29,16 +31,14 @@ else
 end
 
 ruby_icon = rubyw_icon = nil
-[$ruby_name, 'ruby'].each do |i|
+[$ruby_name, 'ruby'].find do |i|
   if i = icons[i]
     ruby_icon = "1 ICON DISCARDABLE "+i.dump+"\n"
-    break
   end
 end
-[$rubyw_name, 'rubyw'].each do |i|
+[$rubyw_name, 'rubyw'].find do |i|
   if i = icons[i]
     rubyw_icon = "1 ICON DISCARDABLE "+i.dump+"\n"
-    break
   end
 end
 dll_icons = []
@@ -46,11 +46,11 @@ icons.keys.sort.each do |i|
   dll_icons << "#{dll_icons.size + 1} ICON DISCARDABLE "+icons[i].dump+"\n"
 end
 
-[ # base name    extension         file type  icons
-  [$ruby_name,   CONFIG["EXEEXT"], 'VFT_APP', ruby_icon],
-  [$rubyw_name,  CONFIG["EXEEXT"], 'VFT_APP', rubyw_icon || ruby_icon],
-  [$so_name,     '.dll',           'VFT_DLL', dll_icons],
-].each do |base, ext, type, icons|
+[ # base name    extension         file type  desc, icons
+  [$ruby_name,   CONFIG["EXEEXT"], 'VFT_APP', 'CUI', ruby_icon],
+  [$rubyw_name,  CONFIG["EXEEXT"], 'VFT_APP', 'GUI', rubyw_icon || ruby_icon],
+  [$so_name,     '.dll',           'VFT_DLL', 'DLL', dll_icons.join],
+].each do |base, ext, type, desc, icons|
   open(base + '.rc', "w") { |f|
     f.binmode if /mingw/ =~ RUBY_PLATFORM
 
@@ -60,7 +60,7 @@ end
 #include <winver.h>
 #endif
 
-#{icons}
+#{icons || ''}
 VS_VERSION_INFO VERSIONINFO
  FILEVERSION    #{fversion}
  PRODUCTVERSION #{fversion}
@@ -74,7 +74,7 @@ BEGIN
  BEGIN
   BLOCK "000004b0"
   BEGIN
-   VALUE "FileDescription",  "Ruby interpreter\\0"
+   VALUE "FileDescription",  "Ruby interpreter (#{desc}) #{rversion} [#{RUBY_PLATFORM}]\\0"
    VALUE "FileVersion",      "#{fversion}\\0"
    VALUE "Home Page",        "http://www.ruby-lang.org/\\0"
    VALUE "InternalName",     "#{base + ext}\\0"
@@ -83,7 +83,7 @@ BEGIN
    VALUE "Platform",         "#{RUBY_PLATFORM}\\0"
    VALUE "ProductVersion",   "#{fversion}\\0"
    VALUE "Release Date",     "#{RUBY_RELEASE_DATE}\\0"
-   VALUE "Version",          "#{RUBY_VERSION}\\0"
+   VALUE "Version",          "#{rversion}\\0"
   END
  END
  BLOCK "VarFileInfo"
